/*
 *  (C) Copyright IX 2004 Michal 'GiM' Spadlinski <gim at skrzynka dot pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License Version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/* this should compile on Solaris, Linux, DragonFlyBSD */
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include <fcntl.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>

#include "data.h"
#include "coolstorage.h"

enum { G_INODE, G_DEVICE, G_RDEVICE, G_BLOCKS, G_RBLOCKS, G_BLKSIZE, G_MODTIME, G_ACCTIME, G_FOLLOW, G_NOERRORS, G_SORT, G_REVERSE, G_RECURSE, G_SYMS, G_HELP, G_ALL };
const char opcje[G_ALL] = "idDbBsmafEorRlh";
const char maval[G_ALL] = "            i  ";
int bit_opcji[G_ALL];
int val_opcji[G_ALL];
int (*fp)(const char *name, struct stat *buf);
const char __fname[] = "lstat()";
const char *fname = __fname;

int usage(void)
{
	printf("usage:\n"
		" gls [-%s] file [file] ...\n"
		" -i - print inode #\n"
		" -d - print device #\n"
		" -D - print rdev # [-d required]\n"
		" -b - size in blk_size blocks\n"
		" -B - size in real blocks [overrides -b]\n"
		" -s - print blocksize\n"
		" -m - print modification time\n"
		" -a - print access time\n"
		" -f - follow symlinks\n"
		" -l - print symlinks' number\n"
		" -E - supress errors\n"
		" -o - sort by name\n"
		" -r - reverse sort order [enables -o]\n"
		" -R [num] - recursive (num is recursion depth) (default: 1)\n"
		" -h - this info\n"
		"If neither -b nor -B is present size in bytes is printed\n"
		"If [num] for -R is not present value of 2 is used\n",
		opcje);
	return 0;
}

int gstrcmp(const void *a, const void *b)
{
	int i = 1-(bit_opcji[G_REVERSE]<<1);
	if (i == 1 && !bit_opcji[G_SORT]) i=0;
	return i*strcmp(*(char **)a, *(char **)b);
}

/* this function is responsible for displaying everything
 */
int fileinfo(char *name)
{
	struct stat st;
	struct passwd *p;
	struct group *g;
	mode_t m, fm;
	int ds;
	time_t mt, at, ct;
	char *t[3], form[30];

	if (fp(name, &st) == -1)
	{
		if (!bit_opcji[G_NOERRORS])
			perror(fname);
		return -1;
	}

	if (bit_opcji[G_INODE])
		printf("%9lu ", st.st_ino); 

	m = st.st_mode;
	switch(fm=(m&S_IFMT))
	{
		case S_IFSOCK:
			fputc('s', stdout); break;
		case S_IFLNK:
			fputc('l', stdout); break;
		case S_IFREG:
			fputc('-', stdout); break;
		case S_IFBLK:
			fputc('b', stdout); break;
		case S_IFDIR:
			fputc('d', stdout); break;
		case S_IFCHR:
			fputc('c', stdout); break;
		case S_IFIFO:
			fputc('p', stdout); break;
		default:				/* unknown */
			fputc('X', stdout); break;
			
	}
	fputc(m&S_IRUSR?'r':'-', stdout);
	fputc(m&S_IWUSR?'w':'-', stdout);
	if (m&S_IXUSR)	fputc(m&S_ISUID?'s':'x', stdout);
	else		fputc(m&S_ISUID?'S':'-', stdout);

	fputc(m&S_IRGRP?'r':'-', stdout);
	fputc(m&S_IWGRP?'w':'-', stdout);
	if (m&S_IXGRP)	fputc(m&S_ISGID?'s':'x', stdout);
	else		fputc(m&S_ISGID?'S':'-', stdout);
	
	fputc(m&S_IROTH?'r':'-', stdout);
	fputc(m&S_IWOTH?'w':'-', stdout);
	if (m&S_IXOTH)	fputc(m&S_ISVTX?'t':'x', stdout);
	else		fputc(m&S_ISVTX?'T':'-', stdout);

	if (bit_opcji[G_SYMS]) 
		printf(" %02lu", st.st_nlink);

	if (bit_opcji[G_DEVICE])
	{
		/*
		 * compatibility fjuczer
		 */
		ds = sizeof(dev_t);
		sprintf(form, " %%0%d%sX %%%su", ds<<1, ds>4?"ll":"l", ds>4?"ll":"l");
		printf(form, st.st_dev, st.st_dev);
		if (bit_opcji[G_RDEVICE] && (fm == S_IFBLK || fm == S_IFCHR))
			printf(form, st.st_rdev, st.st_rdev);
	}

	if ((p = getpwuid(st.st_uid)))
		printf(" %12s", p->pw_name);
	else
		printf(" (%10d)", st.st_uid);

	if ((g = getgrgid(st.st_gid)))
		printf(" %-12s", g->gr_name);
	else
		printf(" (%10d)", st.st_gid);

	if (bit_opcji[G_BLKSIZE])
		printf(" %9ld", st.st_blksize);
		
	if (bit_opcji[G_RBLOCKS])
		printf(" %9ld", st.st_blocks);
	else if (bit_opcji[G_BLOCKS])
		printf(" %9ld", (long int)(st.st_size/(float)st.st_blksize));
	else 
		printf(" %9ld", st.st_size);
	ct = st.st_ctime;
	mt = st.st_mtime;
	at = st.st_atime;
	t[0] = data(ct);
	t[1] = data(mt);
	t[2] = data(at);
	printf(" %s", t[0]);
	if (bit_opcji[G_MODTIME])
		printf(" %s", t[1]);
	if (bit_opcji[G_ACCTIME])
		printf(" %s", t[2]);
	printf(" %s\n", name);
	free(t[0]);
	free(t[1]);
	free(t[2]);
	return 0;
}

/* this is responsible for walking through
 * directories
 * if -R is passed depth is limited
 */
char path[4096] = "";
int cdepth = 0;
int curplen = 0;
int max_depth = 0;

int is_recurse(char *dname)
{
	DIR *d;
	struct stat st;
	struct dirent *de;
	struct strs_t dirz, philez;
	int l, i, total = 0;

	if (lstat(dname, &st) == -1)
		st.st_mode = 0;

	if((st.st_mode & S_IFMT & S_IFLNK) || (d=opendir(dname))==NULL)
	{
		fileinfo(dname);
		return -1;
	}

	strncpy(path, dname, 4095);
	/* strlen(dname) will always be > 0 */
	if (strlen(path)<4095 && path[strlen(path)-1] != '/')
		strcat (path, "/");

	strs_init(&dirz), strs_init(&philez);

	l = strlen(path);
	while ((de = readdir(d)))
	{
		strncat(path, de->d_name, 4095-l);
		if (!lstat(path, &st))
		{
			st.st_mode &= S_IFMT;
			total += st.st_blocks;
			if (bit_opcji[G_SORT] || bit_opcji[G_REVERSE])
			{
				if ((st.st_mode & S_IFDIR) &&
						!(st.st_mode & S_IFLNK))
					strs_add(&dirz, path, 1);
				else 
					strs_add(&philez, path, 1);
			} else {
				fileinfo(path);
				if ((st.st_mode & S_IFDIR) &&
						!(st.st_mode & S_IFLNK))
				{
					cdepth++;
					if (strcmp(de->d_name, ".") &&
						strcmp(de->d_name, ".."))
						gointo(path);
					cdepth--;
				}
			}
		} else {
			if (!bit_opcji[G_NOERRORS])
				perror("lstat");
		}
		*(strrchr(path,'/')+1)='\0';
	}
	closedir(d);

	printf("Total blocks: %9d %s\n", total, dname);
	if (!bit_opcji[G_REVERSE] && !bit_opcji[G_SORT])
		return 0;

	if (dirz.act)
		qsort(dirz.strs, dirz.act, sizeof(char *), gstrcmp);

	if (philez.act)
		qsort(philez.strs, philez.act, sizeof(char *), gstrcmp);

	if (bit_opcji[G_REVERSE])
		for (i=0; i<philez.act; i++)
		{
			fileinfo(philez.strs[i]);
			free(philez.strs[i]);
		}
	
	for (i=0; i<dirz.act; i++)
		fileinfo(dirz.strs[i]);

	if (!bit_opcji[G_REVERSE])
		for (i=0; i<philez.act; i++)
		{
			fileinfo(philez.strs[i]);
			free(philez.strs[i]);
		}
	
	strs_free(&philez, 0);
	
	for (i=0; i<dirz.act; i++)
	{
		if (strcmp(dirz.strs[i]+l, "..") && strcmp(dirz.strs[i]+l, "."))
		{
			strcpy(path, dirz.strs[i]);
			cdepth++;
			gointo(dirz.strs[i]);
			cdepth--;
		}
		free(dirz.strs[i]);
	}
	strs_free(&dirz, 0);
	return 0;

}

int gointo(char *dname)
{
	DIR *d;
	struct stat st;
	struct dirent *de;
	struct strs_t dirz, philez;
	int i, total = 0;
	
	if (cdepth >= max_depth) 
		return 0;

	if (lstat(dname, &st) == -1)
		st.st_mode = 0;

	if((st.st_mode & S_IFMT & S_IFLNK) || (d=opendir(dname))==NULL)
		return -1;

	strs_init(&dirz), strs_init(&philez);

	while((de = readdir(d)))
	{
		strcat (path, "/");
		curplen = 4096-(strlen(path)+10);
		if (curplen < 0) curplen = 0;
		strncat (path, de->d_name, curplen);

		if (!lstat(path, &st)) {
			total += st.st_blocks;
			if (strcmp(de->d_name,"..") && strcmp(de->d_name,"."))
			{
				if (bit_opcji[G_SORT] || bit_opcji[G_REVERSE])
				{
					st.st_mode &= S_IFMT;
					if ((st.st_mode & S_IFDIR) &&
						!(st.st_mode & S_IFLNK))
						strs_add(&dirz, path, 1);
					else 
						strs_add(&philez, path, 1);
				} else {
					fileinfo(path);
					st.st_mode &= S_IFMT;
					if ((st.st_mode & S_IFDIR) &&
						!(st.st_mode & S_IFLNK)) {
						cdepth++;
						gointo(path);
						cdepth--;
					}
				}
			}
		} else
			if (!bit_opcji[G_NOERRORS])
				perror("_stat()\n");
		*strrchr(path,'/')='\0';
	}
	closedir(d);

	printf("Total blocks: %9d %s\n", total, dname);
    if (bit_opcji[G_SORT] || bit_opcji[G_REVERSE])
    {
	if (bit_opcji[G_REVERSE])
		for (i=0; i<philez.act; i++)
		{
			fileinfo(philez.strs[i]);
			free(philez.strs[i]);
		}
	
	for (i=0; i<dirz.act; i++)
		fileinfo(dirz.strs[i]);

	if (!bit_opcji[G_REVERSE])
		for (i=0; i<philez.act; i++)
		{
			fileinfo(philez.strs[i]);
			free(philez.strs[i]);
		}
	
	strs_free(&philez, 0);
	for (i=0; i<dirz.act; i++)
	{
		strcpy(path, dirz.strs[i]);
		cdepth++;
		gointo(dirz.strs[i]);
		cdepth--;
		free(dirz.strs[i]);
	}
	strs_free(&dirz, 0);
	
    }

	return 0;
}

int gatoi(char *buf, int *a)
{
	char *x;
	long t;
	
	t = strtol(buf, &x, 10);
	if (x == buf) return (1);
	*a = t;

	return (0);
}

int main(int argc, char **argv)
{
	int i, j;
	char *op;
	struct stat st;
	struct strs_t dirz, philez;

	if (argc < 2)
	{
		usage();
		return -1;
	}

	/* mini-parser of passwd options... */
	for (i=0; i<G_ALL; i++)
		bit_opcji[i] = val_opcji[i] = 0;

	for (j=1; j<argc; j++)
		if (*argv[j] == '-')
			for (i=1; i<strlen(argv[j]); i++)
				if ((op = strchr(opcje, argv[j][i])))
				{
					bit_opcji[op-opcje] = 1;
					if (maval[op-opcje] != ' ')
						val_opcji[op-opcje] = j+1;
				}

	/* if -h on command line, just display help and disappear */
	if (bit_opcji[G_HELP])
		return usage();

	if (val_opcji[G_RECURSE])
	{
		max_depth = 2;
		if (val_opcji[G_RECURSE]<argc)
			if (gatoi(argv[val_opcji[G_RECURSE]],&max_depth))
				val_opcji[G_RECURSE] = 0;
	}

	if (max_depth > 100) max_depth = 100;
	if (max_depth < 1)   max_depth = 1;

	fp = stat;
	if (bit_opcji[G_FOLLOW]) 
		fname++;
	else
		fp = lstat;
	
	strs_init(&dirz);
	strs_init(&philez);
	for (j=1; j<argc; j++)
		if (*argv[j] != '-')
		{
			/* this is shitty, but I'm lazy
			 */
			for (i=0; i<G_ALL; i++)
				if (val_opcji[i] == j) break;
			if (i == G_ALL) {
				if (!lstat(argv[j], &st))
				{
					st.st_mode &= S_IFMT;
					if ((st.st_mode & S_IFDIR) &&
							!(st.st_mode & S_IFLNK))
						strs_add(&dirz, argv[j], 1);
					else 
						strs_add(&philez, argv[j], 1);
				} else {
					if (!bit_opcji[G_NOERRORS])
						perror("lstat");
				}
			}
		}

	if (dirz.act)
		qsort(dirz.strs, dirz.act, sizeof(char *), gstrcmp);

	if (philez.act)
		qsort(philez.strs, philez.act, sizeof(char *), gstrcmp);

	if (bit_opcji[G_REVERSE])
		for (i=0; i<philez.act; i++)
		{
			fileinfo(philez.strs[i]);
			free(philez.strs[i]);
		}
	
	for (i=0; i<dirz.act; i++)
	{
		is_recurse(dirz.strs[i]);
		free(dirz.strs[i]);
	}
	if (!bit_opcji[G_REVERSE])
		for (i=0; i<philez.act; i++)
		{
			fileinfo(philez.strs[i]);
			free(philez.strs[i]);
		}
	
	strs_free(&dirz, 0);
	strs_free(&philez, 0);
	
	return 0;
}

